/*--------------------------------------------------------------------------*\
 |                                                                          |
 |  This program is free software; you can redistribute it and/or modify    |
 |  it under the terms of the GNU General Public License as published by    |
 |  the Free Software Foundation; either version 2, or (at your option)     |
 |  any later version.                                                      |
 |                                                                          |
 |  This program is distributed in the hope that it will be useful,         |
 |  but WITHOUT ANY WARRANTY; without even the implied warranty of          |
 |  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           |
 |  GNU General Public License for more details.                            |
 |                                                                          |
 |  You should have received a copy of the GNU General Public License       |
 |  along with this program; if not, write to the Free Software             |
 |  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               |
 |                                                                          |
 |  Copyright (C) 2003                                                      |
 |                                                                          |
 |         , __                 , __                                        |
 |        /|/  \               /|/  \                                       |
 |         | __/ _   ,_         | __/ _   ,_                                | 
 |         |   \|/  /  |  |   | |   \|/  /  |  |   |                        |
 |         |(__/|__/   |_/ \_/|/|(__/|__/   |_/ \_/|/                       |
 |                           /|                   /|                        |
 |                           \|                   \|                        |
 |                                                                          |
 |      Enrico Bertolazzi                                                   |
 |      Dipartimento di Ingegneria Meccanica e Strutturale                  |
 |      Universita` degli Studi di Trento                                   |
 |      Via Mesiano 77, I-38050 Trento, Italy                               |
 |      email: enrico.bertolazzi@unitn.it                                   |
 |                                                                          |
\*--------------------------------------------------------------------------*/

/**
 * 
 * @mainpage BLOCK LU
 * @date     May 30, 2006
 * @version  1.0
 * @note     first release May 30, 2006
 *
 * @author   Enrico Bertolazzi
 *
 * @par      Affiliation:
 *           Department of Mechanics and Structures Engineering <br>
 *           University of Trento <br>
 *           via Mesiano 77, I -- 38050 Trento, Italy <br>
 *           enrico.bertolazzi@ing.unitn.it
 *
 * @par Preface
 *
 * @par Abstract
 * 
 */

#ifndef BLOCK_LU_HH
#define BLOCK_LU_HH

#include "Malloc.hh"

#ifndef CHAR_TYPE
#define CHAR_TYPE char
#endif

#ifndef REAL_TYPE
#define REAL_TYPE double
#endif

#ifndef INTEGER_TYPE
#define INTEGER_TYPE int
#endif

namespace BlockLUDefine {

  typedef CHAR_TYPE         charType           ; 
  typedef charType*         charPointer        ;  
  typedef const charType*   charConstPointer   ; 
  typedef charType&         charReference      ;
  typedef const charType&   charConstReference ;

  typedef REAL_TYPE         valueType           ;
  typedef valueType*        valuePointer        ;  
  typedef const valueType*  valueConstPointer   ; 
  typedef valueType&        valueReference      ;
  typedef const valueType&  valueConstReference ;
  
  typedef INTEGER_TYPE      indexType           ;
  typedef indexType*        indexPointer        ;  
  typedef const indexType*  indexConstPointer   ; 
  typedef indexType&        indexReference      ;
  typedef const indexType&  indexConstReference ;

  /*
  //   ____  _            _    _    _   _ 
  //  | __ )| | ___   ___| | _| |  | | | |
  //  |  _ \| |/ _ \ / __| |/ / |  | | | |
  //  | |_) | | (_) | (__|   <| |__| |_| |
  //  |____/|_|\___/ \___|_|\_\_____\___/ 
  //
  */

  class BlockLU {

    Malloc<valueType> baseValue ;
    Malloc<indexType> baseIndex ;

    BlockLU(BlockLU const &) ;
    BlockLU const & operator = (BlockLU const &) ;

    indexType nblock ; //!< total number of blocks
    indexType n      ; //!< size of square blocks
    indexType m      ; //!< number final rows (m>=n)
    indexType N      ; //!< n * (nblock+1) + q
    indexType nnz    ; //!< total number of non zeros

    /*
    //
    //  Matrix structure
    //
    //                  n * nblock
    //    ________________^_________________
    //   /                                  \
    //     n     n     n                        n     q
    //  +-----+-----+-----+----.........-----+-----+-----+    \
    //  |  Ad | Au  |  0  |                  |  0  |  0  | n   |
    //  +-----+-----+-----+             -----+-----+-----+     |
    //  |  0  | Ad  | Au  |  0               |  0  |  0  | n   |
    //  +-----+-----+-----+-----+       -----+-----+-----+     |
    //  |  0  |  0  | Ad  | Au  |            |  0  |  0  | n   |
    //  +-----+-----+-----+-----+       -----+-----+-----+     |   
    //  |                                                :     |
    //  :                                                :      > n * nblock
    //  :                                                :     | 
    //  :                                                :     |
    //  :                                                :     |
    //  :                              +-----+-----+-----+     |
    //  :                              | Au  |  0  |  0  |     |
    //  :                        +-----+-----+-----+-----+     |
    //  :                        |  0  | Ad  | Au  |  0  | n   |
    //  +-----+-----+---......---+-----+-----+=====+=====+    /
    //  |     |     |            |     |     !     |     !
    //  | H0  |  0  |            |     |  0  ! HN  | Hq  ! m
    //  |     |     |            |     |     !     |     !
    //  +-----+-----+---......---+-----+-----+=====+=====+
    //
    */
    valueConstPointer AdAu ;
    valueConstPointer H0   ;
    valueConstPointer HN   ;
    valueConstPointer Hq   ;
    
    /*
    //
    //  Working block AdH_blk [ size = nblock * ( n * (n+m) ) ]
    //
    //                 n * nblock
    //    ________________^_________________
    //   /                                  \
    //     n     n     n                      
    //  +-----+-----+-----+----........+-----+
    //  |  Ad | Ad  | Ad  |               Ad | n
    //  +-----+-----+-----+            +-----+
    //  |     |     |            |     |     !
    //  | H0  |  0  |            |     |  0  ! m
    //  |     |     |            |     |     !
    //  +-----+-----+---......---+-----+-----+
    //
    */
    valuePointer AdH_blk ;

    /*
    //
    //  Working block Au_blk [ size = nblock * (n*n) ]
    //
    //                  n * nblock + q
    //    ________________^________________________
    //   /                                         \
    //     n     n     n                  n      q
    //  +-----+-----+-----+----.........-----+------+
    //  |  Au | Au  | Au  |               Au | fill | n
    //  +-----+-----+-----+----.........-----+------+
    //
    */
    valuePointer Au_blk ;

    /*
    //
    //  Working block DD_blk [ size = m * m ]
    //
    //     n     q     (n+q=m)
    //  +=====+=====+
    //  !     |     !
    //  ! HN  | Hq  ! m
    //  !     |     !
    //  +=====+=====+
    //
    */    
    valuePointer DD_blk ;
    
    /*
    //
    //  Working block FF_blk [ size = nblock * (n*m) ]
    //
    //     n     q        (n+q=m)
    //  +-----+-----+   \
    //  |fill |fill | n |
    //  +-----+-----+   |
    //  |fill |fill | n |
    //  +-----+-----+   |
    //  |fill |fill | n |
    //  +-----+-----+   :
    //  |           |   :  n * (nblock-1)
    //  :           :   :
    //  :           :   :
    //  :           :   :
    //  :           :   :
    //  :           :   :
    //  |fill |fill | n |
    //  +-----+-----+   /
    //
    */
    valuePointer FF_blk ;

    // pivot vector
    indexPointer ipiv_blk ;

    void allocate( indexType nblk, indexType n, indexType q ) ;

  public:

    explicit BlockLU();

    ~BlockLU() ;

    void load( indexType         nblk,
               indexType         n,
               indexType         q,
               valueConstPointer AdAu,
               valueConstPointer H0,
               valueConstPointer HN,
               valueConstPointer Hq ) ;

    void amxpy( valueType a, valuePointer x, valuePointer y ) const ;

    indexType factorize() ;

    void solve( valuePointer in_out ) ;

  } ;

}

namespace BlockLULoad {
  using BlockLUDefine::BlockLU ;
}

#endif
