#include "BlockLU.hh"
#include "AmodioLU.hh"
//#include "ArcecoLU.hh"
#include "Alglin.hh"
#include "timing.hh"

#include <cmath>
#include <iostream>
#include <iomanip>

using namespace std ;

/*
// +                      .         +
// | n*n n*n              .         |
// |     n*n n*n          .         |
// |        n*n n*n       .         |
// |                      .         |
// |                      .         |
// |                      . . . . . |
// |                  n*n . n*n n*q |     q = m-n
// | m*n                  . m*n   0 |
// +                                +
//
*/

using namespace BlockLULoad ;
using namespace AmodioLULoad ;
//using namespace ArcecoLULoad ;

typedef BlockLUDefine::valueType valueType ;
typedef BlockLUDefine::indexType indexType ;

int
main() {
  Timing    tm ;
  BlockLU   bm ;
  AmodioLU  am ;
  //ArcecoLU bm ;
  bool      ok ;
  valueType elapsed ;

  indexType   nblk = 1700 ;
  indexType   n    = 16   ;
  indexType   q    = 9    ;
  indexType   m    = n+q  ;
  indexType   nxn  = n*n  ;
  indexType   N    = (nblk+1)*n+q ;

  valueType * AdAu = new valueType[2*nblk*nxn] ;
  valueType * Hi   = new valueType[m*n] ;
  valueType * Hf   = new valueType[m*n] ;
  valueType * Hmu  = new valueType[m*q] ;
  valueType * xb   = new valueType[N] ;
  valueType * xb1  = new valueType[N] ;
  valueType * x    = new valueType[N] ;
  
  cout << setw(12) << "factor" ; // 1
  cout << setw(12) << "solve"  ; // 1
  cout << setw(12) << "err"    ; // 1
  cout << setw(12) << "factor (Amo)" ; // 1
  cout << setw(12) << "solve (Amo)"  ; // 1
  cout << setw(12) << "err"    ; // 1
  cout << '\n' ;

  for ( indexType kkk = 0 ; kkk < 100 ; ++kkk ) {
  
    // riempie la matrice
    for ( indexType i = 0 ; i < n ; ++i ) {
      for ( indexType j = 0 ; j < n ; ++j ) {
        indexType iaddr = i+j*n ;
        for ( indexType k = 0 ; k < nblk ; ++k ) {
          valueType * Ad = AdAu + 2*k*nxn ;
          valueType * Au = Ad   + nxn ;
          Ad[iaddr] = (rand()%1000) ;
          Au[iaddr] = (rand()%2000) ;
        }
      }
    }
    // riempie la matrice
    for ( indexType i = 0 ; i < m ; ++i ) {
      for ( indexType j = 0 ; j < n ; ++j ) {
        indexType iaddr = i+j*m ;
        Hi[iaddr] = (rand()%1000) ;
        Hf[iaddr] = (rand()%1000) ;
      }
    }
    // riempie la matrice
    for ( indexType i = 0 ; i < m ; ++i ) {
      for ( indexType j = 0 ; j < q ; ++j ) {
        indexType iaddr = i+j*m ;
        Hmu[iaddr] = (rand()%1000) ;
      }
    }

    //
    for ( indexType i = 0 ; i < N ; ++i ) {
      x[i]   = i ; // (arc4random()%234) ;
      xb[i]  = 0 ;
      xb1[i] = 0 ;
    }

    //////////////////////////////
    bm . load( nblk, n, q, AdAu, Hi, Hf, Hmu ) ;
    am . load( nblk, n, q, AdAu, Hi, Hf, Hmu ) ;    
    bm . amxpy( 1, x, xb  ) ;
    bm . amxpy( 1, x, xb1 ) ;

    tm . start() ;
    ok = bm . factorize() ; if ( !ok ) cout << "factorize error\n" ;
    elapsed = tm . milliseconds() ;
    cout << setw(10) << elapsed << "ms" ;

    tm . start() ;
    bm . solve( xb ) ;
    elapsed = tm . milliseconds() ;
    cout << setw(10) << elapsed << "ms" ;
    
    alglin::axpy( N, -1.0, x, 1, xb, 1 ) ;
    cout << "  " << setw(10) << alglin::nrm2( N, xb, 1 ) ;

    tm . start() ;
    indexType ierr = am . factorize() ;
    if ( ierr ) cout << "factorize error ierr = " << ierr << '\n' ;
    elapsed = tm . milliseconds() ;
    cout << setw(10) << elapsed << "ms" ;

    tm . start() ;
    am . solve( xb1 ) ;
    elapsed = tm . milliseconds() ;
    cout << setw(10) << elapsed << "ms" ;
    alglin::axpy( N, -1.0, x, 1, xb1, 1 ) ;    
    cout << "  " << setw(10) << alglin::nrm2( N, xb1, 1 ) ;
    cout << '\n' ;

  }

}
